import ast
from typing import Optional

from cached_property import cached_property

from projectreport.analyzer.parsers.python.base import PythonParser
from projectreport.version import Version


class PythonSetupPyParser(PythonParser):
    @cached_property
    def docstring(self) -> Optional[str]:
        if self.parsed is None:
            return None

        # Walk ast to look for a description kwarg being passed to the setup function.
        # Extract any string that is passed to it
        for node in ast.walk(self.parsed):
            if isinstance(node, ast.keyword) and node.arg == "description":
                if isinstance(node.value, ast.Str):
                    return node.value.s
                elif isinstance(node.value, ast.NameConstant):
                    return node.value.value
        return None

    @cached_property
    def version(self) -> Optional[Version]:
        if self.parsed is None:
            return None
        # Walk ast to look for a version kwarg being passed to the setup function.
        # Extract any string or number that is passed to it
        for node in ast.walk(self.parsed):
            if isinstance(node, ast.keyword) and node.arg == "version":
                if isinstance(node.value, ast.Str):
                    return Version.from_str(node.value.s)
                elif isinstance(node.value, ast.Num):
                    return Version.from_str(str(node.value.n))
        return None



import os
from pathlib import Path
from typing import Dict, Optional, Sequence

from cached_property import cached_property

from projectreport.analyzer.parsers.base import Parser
from projectreport.analyzer.parsers.index import PARSER_DOC_FILES
from projectreport.version import Version


class MultiFileParser(Parser):
    def __init__(
        self,
        path: str,
        file_names: Sequence[str],
        file_parsers: Optional[Dict[str, Parser]] = None,
    ):
        """
        :param path: This should be the path of a folder, rather than a path to a file
          that the other parsers accept.
        :param file_parsers: Defaults to PARSER_DOC_FILES.
        """
        self.file_parsers = file_parsers or PARSER_DOC_FILES
        self.file_names = file_names
        super().__init__(path)

    @cached_property
    def has_any_match(self) -> bool:
        for file, parser in self.file_parsers.items():
            if file in self.file_names:
                return True
        return False

    @cached_property
    def docstring(self) -> Optional[str]:
        return self._get_attr_from_first_parser_to_return_non_none("docstring")

    @cached_property
    def version(self) -> Optional[Version]:
        return self._get_attr_from_first_parser_to_return_non_none("version")

    def _get_attr_from_first_parser_to_return_non_none(self, attr: str):
        for file, parser in self.file_parsers.items():
            if file in self.file_names:
                full_path = os.path.join(self.path, file)
                parser = PARSER_DOC_FILES[file](full_path)
                value = getattr(parser, attr)
                if value is not None:
                    return value
        return None

    @cached_property
    def parsed(self):
        raise NotImplementedError

    @cached_property
    def contents(self) -> str:
        raise NotImplementedError

    @cached_property
    def file_name(self) -> str:
        raise NotImplementedError



from projectreport.analyzer.parsers.python.init import PythonInitParser
from projectreport.analyzer.parsers.python.setup_cfg import PythonSetupCfgParser
from projectreport.analyzer.parsers.python.setup_py import PythonSetupPyParser
from projectreport.version import Version
from tests.config import (
    PYTHON_DUNDER_VERSION_PROJECT_PATH,
    PYTHON_SETUP_CFG_VERSION_PROJECT_PATH,
    PYTHON_SETUP_PY_VERSION_PROJECT_PATH,
)


def test_init_py_parser():
    input_file = PYTHON_DUNDER_VERSION_PROJECT_PATH / "__init__.py"
    parser = PythonInitParser(input_file)
    assert parser.docstring == "An example Python package for testing purposes"
    assert parser.version == Version.from_str("0.0.1")


def test_setup_py_parser():
    input_file = PYTHON_SETUP_PY_VERSION_PROJECT_PATH / "setup.py"
    parser = PythonSetupPyParser(input_file)
    assert (
        parser.docstring
        == "[from setup.py] An example Python package for testing purposes (version from setup.py)"
    )
    assert parser.version == Version.from_str("0.0.1")


def test_setup_cfg_parser():
    input_file = PYTHON_SETUP_CFG_VERSION_PROJECT_PATH / "setup.cfg"
    parser = PythonSetupCfgParser(input_file)
    assert (
        parser.docstring
        == "[from setup.cfg] An example Python package for testing purposes (version from setup.cfg)"
    )
    assert parser.version == Version.from_str("1.0.0")



import ast
from typing import Optional

from cached_property import cached_property

from projectreport.analyzer.parsers.python.base import PythonParser
from projectreport.version import Version


class PythonInitParser(PythonParser):
    @cached_property
    def docstring(self) -> Optional[str]:
        if self.parsed is None:
            return None
        return ast.get_docstring(self.parsed)

    @cached_property
    def version(self) -> Optional[Version]:
        if self.parsed is None:
            return None
        # Walk ast to look for __version__ variable. If it is defined, extract the version from it
        for node in ast.walk(self.parsed):
            if isinstance(node, ast.Assign) and node.targets[0].id == "__version__":  # type: ignore
                # Extract version from __version__ = "1.2.3"
                if isinstance(node.value, ast.Str):
                    return Version.from_str(node.value.s)
                # Extract version from __version__ = 1.2
                elif isinstance(node.value, ast.Num):
                    return Version.from_str(str(node.value.n))
        return None
